theory Typing
imports Syntax

begin

datatype Type = Type
datatype Term = Term "ty"
types tm_env = "(name * Term) set"
types ty_env = "(name * Type) set"

syntax
  var_term :: "name \<Rightarrow> ty \<Rightarrow> (name * Term)"  ("(_\<rhd>_)" [90,90] 90)
  var_type :: "name \<Rightarrow> (name * Type)"  ("(_\<rhd>'\<star>)" [90] 90)
translations
  "(x\<rhd>\<tau>)" \<rightleftharpoons> "(x, Term \<tau>)"
  "(x\<rhd>\<star>)" \<rightleftharpoons> "(x, Type)"

syntax
  addbinding :: "(name * Term) set \<Rightarrow> name \<Rightarrow> ty \<Rightarrow> (name * Term) set"
    ("_','(_':_')" [90,0,0] 91)
  addbinding_ty :: "(name * Type) set \<Rightarrow> name \<Rightarrow> (name * Type) set"
    ("_','(_':'\<star>')" [90,0] 91)
translations
  "\<Gamma>,(x:\<tau>)" \<rightleftharpoons> "insert (x\<rhd>\<tau>) \<Gamma>"
  "\<Delta>,(x:\<star>)" \<rightleftharpoons> "insert (x\<rhd>\<star>) \<Delta>"

lemma only_term:
  assumes "\<Gamma> defines x"
  shows "\<exists>\<tau>. (x\<rhd>\<tau>) \<in> \<Gamma>"
proof -
  obtain b where "(x, b) \<in> \<Gamma>" using prems defines_sanity by blast
  thus ?thesis by (cases b, auto)
qed

constdefs free_ty_env :: "tm_env \<Rightarrow> name set"
  "free_ty_env \<Gamma> \<equiv> \<Union> { set (free_ty \<tau>) | x \<tau>. (x \<rhd> \<tau>) \<in> \<Gamma> }"

constdefs free_db_ty_env :: "tm_env \<Rightarrow> name set"
  "free_db_ty_env \<Gamma> \<equiv> \<Union> { set (free_db_ty \<tau>) | x \<tau>. (x \<rhd> \<tau>) \<in> \<Gamma> }"

syntax
  is_free_in_env :: "name \<Rightarrow> tm_env \<Rightarrow> bool"  ("_ free in env _" [900,900] 91)
translations
  "a free in env \<Gamma>" \<rightleftharpoons> "a \<in> free_ty_env \<Gamma>"

lemma free_env_finite:
  assumes "finite \<Gamma>"
  shows "finite (free_ty_env \<Gamma>)"
proof -
  let ?f = "%m. case m of (Pair x b) \<Rightarrow>
                  case b of (Term \<tau>) \<Rightarrow>
                    set (free_ty \<tau>)"
  let ?A = "{F. \<exists>x \<tau>. F = set (free_ty \<tau>) \<and> (x\<rhd>\<tau>) \<in> \<Gamma>}"
  have "finite (\<Union>{F. \<exists>x \<tau>. F = set (free_ty \<tau>) \<and> (x\<rhd>\<tau>) \<in> \<Gamma>})"
    using prems finite_surj[of \<Gamma> ?A ?f] by force
  thus ?thesis using prems finite_Union by (auto simp add: free_ty_env_def)
qed

lemma free_ty_env_subset:
  assumes "\<not> x free in env \<Gamma>"
  and "\<Gamma>' \<subseteq> \<Gamma>"
  shows "\<not> x free in env \<Gamma>'"
using prems by (auto simp add: free_ty_env_def)

constdefs vsubst_ty_env:: "ty \<Rightarrow> nat \<Rightarrow> tm_env \<Rightarrow> tm_env"
  "vsubst_ty_env s i \<Gamma> \<equiv> { (x\<rhd>vsubst_ty s i \<tau>) | x \<tau>. (x \<rhd> \<tau>) \<in> \<Gamma> }"

constdefs convert_ty_env:: "nat \<Rightarrow> nat \<Rightarrow> tm_env \<Rightarrow> tm_env"
  "convert_ty_env j i \<Gamma> \<equiv> { (x\<rhd>convert_ty j i \<tau>) | x \<tau>. (x \<rhd> \<tau>) \<in> \<Gamma> }"

constdefs vsubst_ty_func :: "ty \<Rightarrow> nat \<Rightarrow> (name * Term) \<Rightarrow> (name * Term)"
  "vsubst_ty_func s i m \<equiv> case m of (Pair x b) \<Rightarrow>
                            case b of (Term \<tau>) \<Rightarrow>
                              (x\<rhd>vsubst_ty s i \<tau>)"

lemma vsubst_ty_env_sanity:
  shows "(vsubst_ty_func s i) ` \<Gamma> = vsubst_ty_env s i \<Gamma>"
proof (auto simp add: vsubst_ty_env_def vsubst_ty_func_def)
  fix a b aa ba
  assume "(a, b) = (case ba of Term \<tau> \<Rightarrow> (aa\<rhd>(vsubst_ty s i \<tau>)))"
    and "(aa, ba) \<in> \<Gamma>"
  show "\<exists>x \<tau>. (case ba of Term \<tau> \<Rightarrow> (aa\<rhd>(vsubst_ty s i \<tau>))) = (x\<rhd>(vsubst_ty s i \<tau>))
            \<and> (x\<rhd>\<tau>) \<in> \<Gamma>" using prems by (cases ba, auto)
next
  fix x \<tau>
  assume "(x\<rhd>\<tau>) \<in> \<Gamma>"
  show "(x\<rhd>(vsubst_ty s i \<tau>)) \<in> (vsubst_ty_func s i) ` \<Gamma>" proof
    show "(x\<rhd>\<tau>) \<in> \<Gamma>" .
  next
    show "(x\<rhd>(vsubst_ty s i \<tau>)) = vsubst_ty_func s i (x\<rhd>\<tau>)"
      by (auto simp add: vsubst_ty_func_def)
  qed
qed

lemma vsubst_ty_env_dom_sanity:
  shows "dom \<Gamma> = dom (vsubst_ty_env s i \<Gamma>)"
proof (auto simp add: dom_def vsubst_ty_env_def)
  fix x b
  assume "(x, b) \<in> \<Gamma>"
  hence "\<Gamma> defines x" by (auto simp add: dom_def)
  thus "\<exists>\<tau>. (x\<rhd>\<tau>) \<in> \<Gamma>" by (rule only_term)
qed

constdefs
  wf_env_ty :: "ty_env \<Rightarrow> bool" ("_ \<Turnstile> TY")
  "wf_env_ty \<Delta> \<equiv> finite \<Delta>"

consts WFType :: "(ty_env * ty) set"
syntax
  "_wfType" :: "ty_env \<Rightarrow> ty \<Rightarrow> bool"      ("_ \<Turnstile> _" [70,70] 50)
translations
  "\<Delta> \<Turnstile> \<tau>" \<rightleftharpoons> "(\<Delta>, \<tau>) \<in> WFType"

inductive WFType
  intros
  wf_unit[intro!]   : "\<lbrakk> \<Delta> \<Turnstile> TY \<rbrakk> \<Longrightarrow> \<Delta> \<Turnstile> TyUnit"
  wf_base[intro!]   : "\<lbrakk> \<Delta> \<Turnstile> TY \<rbrakk> \<Longrightarrow> \<Delta> \<Turnstile> TyBase n"
  wf_tvar[intro!]   : "\<lbrakk> (x\<rhd>\<star>) \<in> \<Delta>; \<Delta> \<Turnstile> TY \<rbrakk> \<Longrightarrow> \<Delta> \<Turnstile> TyFreeVar x"
  wf_arrow[intro!]  : "\<lbrakk> \<Delta> \<Turnstile> \<tau>1; \<Delta> \<Turnstile> \<tau>2 \<rbrakk> \<Longrightarrow> \<Delta> \<Turnstile> \<tau>1 \<rightarrow> \<tau>2"
  wf_forall[intro!] : "\<lbrakk> \<Delta> \<Turnstile> TY; finite L;
                         \<forall>a. \<not> \<Delta> defines a \<and> \<not> a free in type \<tau> \<and> a \<notin> L \<longrightarrow>
                         \<Delta>,(a:\<star>) \<Turnstile> freshen_ty a \<tau> \<rbrakk> \<Longrightarrow> \<Delta> \<Turnstile> \<forall>.\<tau>"
  wf_ref[intro!]:     "\<lbrakk> \<Delta> \<Turnstile> \<tau> \<rbrakk> \<Longrightarrow> \<Delta> \<Turnstile> TyRef \<tau>"

inductive_cases wf_ty_unit [elim!]:   "\<Delta> \<Turnstile> TyUnit"
inductive_cases wf_ty_base [elim!]:   "\<Delta> \<Turnstile> TyBase n"
inductive_cases wf_ty_tvar [elim!]:   "\<Delta> \<Turnstile> TyFreeVar x"
inductive_cases wf_ty_tfvar [elim!]:   "\<Delta> \<Turnstile> TyVar x"
inductive_cases wf_ty_arrow [elim!]:  "\<Delta> \<Turnstile> \<tau>1 \<rightarrow> \<tau>2"
inductive_cases wf_ty_forall [elim!]: "\<Delta> \<Turnstile> \<forall>.\<tau>"
inductive_cases wf_ty_ref [elim!]:    "\<Delta> \<Turnstile> TyRef \<tau>"


constdefs
  wf_env :: "tm_env \<Rightarrow> ty_env \<Rightarrow> bool" ("_;_ \<Turnstile> OK" [70,70] 50)
  "wf_env \<Gamma> \<Delta> \<equiv> finite \<Gamma> \<and> wf_env_ty \<Delta> \<and> (\<forall>x. \<forall>\<tau>1. \<forall>\<tau>2. (x\<rhd>\<tau>1) \<in> \<Gamma> \<and> (x\<rhd>\<tau>2) \<in> \<Gamma> \<longrightarrow> \<tau>1 = \<tau>2 \<and> \<Delta> \<Turnstile> \<tau>1)"

lemma wf_type_wf_env_ty:
  assumes "\<Delta> \<Turnstile> \<tau>"
  shows "\<Delta> \<Turnstile> TY"
using prems by (induct, auto)

lemma wf_env_wf_env_ty:
  assumes "\<Gamma>;\<Delta> \<Turnstile> OK"
  shows "\<Delta> \<Turnstile> TY"
using prems wf_env_def by auto

consts Typing :: "(tm_env * ty_env * tm * ty) set"
syntax
  "_Typing" :: "tm_env \<Rightarrow> ty_env \<Rightarrow> tm \<Rightarrow> ty \<Rightarrow> bool"   ("_';_ \<turnstile> _ : _" [70,70,70,70] 50)
translations
  "\<Gamma>;\<Delta> \<turnstile> e : \<tau>" \<rightleftharpoons> "(\<Gamma>, \<Delta>, e, \<tau>) \<in> Typing"

inductive Typing
  intros
  t_unit[intro!]   : "\<lbrakk> \<Gamma>;\<Delta> \<Turnstile> OK \<rbrakk> \<Longrightarrow> \<Gamma>;\<Delta> \<turnstile> TmUnit : TyUnit"
  t_var[intro!]    : "\<lbrakk> \<Gamma>;\<Delta> \<Turnstile> OK; (x\<rhd>\<tau>) \<in> \<Gamma> \<rbrakk> \<Longrightarrow> \<Gamma>;\<Delta> \<turnstile> TmFreeVar x : \<tau>"
  t_abs[intro!]    : "\<lbrakk> \<Gamma>;\<Delta> \<Turnstile> OK; finite L; \<Delta> \<Turnstile> \<tau>1;
                        \<forall>x. \<not> x free in e \<and> x \<notin> L \<and> \<not> \<Gamma> defines x \<longrightarrow>
                          \<Gamma>,(x:\<tau>1);\<Delta> \<turnstile> freshen_tm x e : \<tau>2 \<rbrakk> \<Longrightarrow>
                        \<Gamma>;\<Delta> \<turnstile> \<lambda>[\<tau>1]. e : \<tau>1\<rightarrow>\<tau>2"
  t_app[intro!]    : "\<lbrakk> \<Gamma>;\<Delta> \<turnstile> e1 : \<tau>1\<rightarrow>\<tau>2; \<Gamma>;\<Delta> \<turnstile> e2 : \<tau>1 \<rbrakk> \<Longrightarrow>
                        \<Gamma>;\<Delta> \<turnstile> e1\<cdot>e2 : \<tau>2"
  t_poly[intro!]   : "\<lbrakk> \<Gamma>;\<Delta> \<Turnstile> OK; finite L; value v;
                        \<forall>a. \<not> a free in term v \<and> \<not> a free in type \<tau> \<and> a \<notin> L \<and>
                            \<not> a free in env \<Gamma> \<and> \<not> \<Delta> defines a \<longrightarrow>
                          \<Gamma>;\<Delta>,(a:\<star>) \<turnstile> freshen_tmty a v : freshen_ty a \<tau> \<rbrakk> \<Longrightarrow>
                        \<Gamma>;\<Delta> \<turnstile> \<Lambda>[*].v : \<forall>.\<tau>"
  t_tyapp[intro!]  : "\<lbrakk> \<Gamma>;\<Delta> \<turnstile> e1 : \<forall>.\<tau>1; \<Delta> \<Turnstile> \<tau> \<rbrakk> \<Longrightarrow> \<Gamma>;\<Delta> \<turnstile> e1\<langle>\<tau>\<rangle> : vsubst_ty \<tau> 0 \<tau>1"
  t_new[intro!]    : "\<lbrakk> \<Gamma>;\<Delta> \<turnstile> e : \<tau> \<rbrakk> \<Longrightarrow> \<Gamma>;\<Delta> \<turnstile> TmNew e : TyRef \<tau>"
  t_assign[intro!] : "\<lbrakk> \<Gamma>;\<Delta> \<turnstile> e1 : TyRef \<tau>; \<Gamma>;\<Delta> \<turnstile> e2 : \<tau> \<rbrakk> \<Longrightarrow>
                        \<Gamma>;\<Delta> \<turnstile> e1 := e2 : TyUnit"
  t_deref[intro!]  : "\<lbrakk> \<Gamma>;\<Delta> \<turnstile> e : TyRef \<tau> \<rbrakk> \<Longrightarrow> \<Gamma>;\<Delta> \<turnstile> TmDeref e : \<tau>"
  t_loc[intro!]    : "\<lbrakk> \<Gamma>;\<Delta> \<Turnstile> OK; (l\<rhd>\<tau>) \<in> \<Gamma> \<rbrakk> \<Longrightarrow> \<Gamma>;\<Delta> \<turnstile> TmLoc l : TyRef \<tau>"

inductive_cases typing_unit_elim [elim!]:    "\<Gamma>;\<Delta> \<turnstile> TmUnit : \<tau>"
inductive_cases typing_var_elim [elim!]:     "\<Gamma>;\<Delta> \<turnstile> TmVar x : \<tau>"
inductive_cases typing_freevar_elim [elim!]: "\<Gamma>;\<Delta> \<turnstile> TmFreeVar x : \<tau>"
inductive_cases typing_abs_elim [elim!]:     "\<Gamma>;\<Delta> \<turnstile> \<lambda>[\<tau>1]. e : \<tau>"
inductive_cases typing_poly_elim [elim!]:    "\<Gamma>;\<Delta> \<turnstile> \<Lambda>[*]. e : \<tau>"
inductive_cases typing_app_elim [elim!]:     "\<Gamma>;\<Delta> \<turnstile> e1\<cdot>e2 : \<tau>"
inductive_cases typing_app_ty_elim [elim!]:  "\<Gamma>;\<Delta> \<turnstile> e\<langle>\<tau>\<rangle> : \<tau>'"
inductive_cases typing_new_elim [elim!]:     "\<Gamma>;\<Delta> \<turnstile> TmNew e : \<tau>"
inductive_cases typing_assign_elim [elim!]:  "\<Gamma>;\<Delta> \<turnstile> e1 := e2 : \<tau>"
inductive_cases typing_deref_elim [elim!]:   "\<Gamma>;\<Delta> \<turnstile> TmDeref e : \<tau>"
inductive_cases typing_loc_elim [elim!]:     "\<Gamma>;\<Delta> \<turnstile> TmLoc l : \<tau>"

lemma vsubst_tmty_free_db:
  shows "free_db_tm e = free_db_tm (vsubst_tmty \<tau> i e)"
by (induct e arbitrary: i, auto)

lemma wf_env_weakening:
  assumes "\<Gamma>;\<Delta> \<Turnstile> OK"
  and "\<not> \<Gamma> defines x"
  and "\<Delta> \<Turnstile> \<tau>"
  shows "\<Gamma>,(x:\<tau>);\<Delta> \<Turnstile> OK"
using prems by (auto simp add: dom_def wf_env_def)

lemma wf_env_strengthening_ty:
  assumes "\<Delta>,(a:\<star>) \<Turnstile> TY"
  shows "\<Delta> \<Turnstile> TY"
using prems wf_env_ty_def by auto

lemma wf_ty_ok_ty:
  assumes "\<Gamma>;\<Delta> \<Turnstile> OK"
  and "(x\<rhd>\<tau>) \<in> \<Gamma>"
  shows "\<Delta> \<Turnstile> \<tau>"
using prems wf_env_def[of \<Gamma> \<Delta>] by auto

lemma wf_env_fin:
  assumes "\<Gamma>;\<Delta> \<Turnstile> OK"
  shows "finite (dom \<Gamma>)"
using dom_finite wf_env_def[of \<Gamma> \<Delta>] prems by auto

lemma wf_env_ty_fin:
  assumes "\<Gamma>;\<Delta> \<Turnstile> OK"
  shows "finite (dom \<Delta>)"
using wf_env_ty_def[of \<Delta>] wf_env_def[of \<Gamma> \<Delta>] dom_finite prems by auto

lemma wf_env_subset:
  assumes "\<Gamma>;\<Delta> \<Turnstile> OK"
  and Hs: "\<Gamma>' \<subseteq> \<Gamma>"
  shows "\<Gamma>';\<Delta> \<Turnstile> OK"
proof -
  have H1: "finite \<Gamma>"
    and H2: "\<forall>x. \<forall>\<tau>1. \<forall>\<tau>2. (x\<rhd>\<tau>1) \<in> \<Gamma> \<and> (x\<rhd>\<tau>2) \<in> \<Gamma> \<longrightarrow> \<tau>1 = \<tau>2 \<and> \<Delta> \<Turnstile> \<tau>1"
    using prems by (auto simp add: wf_env_def)
  from H1 have "finite \<Gamma>'" using prems by (auto simp add: finite_subset)
  moreover
  have "\<Delta> \<Turnstile> TY" using prems wf_env_wf_env_ty by auto
  moreover
  have "\<forall>x. \<forall>\<tau>1. \<forall>\<tau>2. (x\<rhd>\<tau>1) \<in> \<Gamma>' \<and> (x\<rhd>\<tau>2) \<in> \<Gamma>' \<longrightarrow> \<tau>1 = \<tau>2 \<and> \<Delta> \<Turnstile> \<tau>1" proof auto
    fix x \<tau>1 \<tau>2
    assume "(x\<rhd>\<tau>1) \<in> \<Gamma>'"
      and "(x\<rhd>\<tau>2) \<in> \<Gamma>'"
    hence "(x\<rhd>\<tau>1) \<in> \<Gamma>" and "(x\<rhd>\<tau>2) \<in> \<Gamma>" using prems by auto
    thus "\<tau>1 = \<tau>2" using H2 by auto
  next
    fix x \<tau>1 \<tau>2
    assume "(x\<rhd>\<tau>1) \<in> \<Gamma>'"
      and "(x\<rhd>\<tau>2) \<in> \<Gamma>'"
    hence "(x\<rhd>\<tau>1) \<in> \<Gamma>" and "(x\<rhd>\<tau>2) \<in> \<Gamma>" using prems by auto
    thus "\<Delta> \<Turnstile> \<tau>1" using H2 by auto
  qed
  ultimately
  show ?thesis by (auto simp add: wf_env_def) 
qed

lemma ok_weakening_ty:
  assumes "\<Delta> \<Turnstile> \<tau>"
  shows "\<Delta>,(z:\<star>) \<Turnstile> \<tau>"
proof -
  have H: "\<And>\<Delta>' .\<Delta>' \<Turnstile> TY \<Longrightarrow> \<Delta>',(z:\<star>) \<Turnstile> TY" using wf_env_ty_def by auto
  show ?thesis using prems H by (induct, auto simp add: env_permute)
qed

lemma ok_weakening_tyenv:
  assumes "\<Gamma>;\<Delta> \<Turnstile> OK"
  and "\<not> \<Delta> defines z"
  shows "\<Gamma>;\<Delta>,(a:\<star>) \<Turnstile> OK"
proof -
have "\<Delta>,(a:\<star>) \<Turnstile> TY" using wf_env_wf_env_ty wf_env_ty_def prems by auto
moreover
have "(\<forall>x \<tau>1 \<tau>2. x\<rhd>\<tau>1 \<in> \<Gamma> \<and> x\<rhd>\<tau>2 \<in> \<Gamma> \<longrightarrow> \<tau>1 = \<tau>2 \<and> \<Delta> \<Turnstile> \<tau>1)" 
  using wf_env_def prems by auto
hence "(\<forall>x \<tau>1 \<tau>2. x\<rhd>\<tau>1 \<in> \<Gamma> \<and> x\<rhd>\<tau>2 \<in> \<Gamma> \<longrightarrow> \<tau>1 = \<tau>2 \<and> \<Delta>,(a:\<star>) \<Turnstile> \<tau>1)"
  using ok_weakening_ty by auto
ultimately
show ?thesis using prems wf_env_def by auto
qed

lemma wf_env_elim_ty:
  assumes "\<Gamma>;\<Delta> \<Turnstile> OK"
  and "(x\<rhd>\<tau>) \<in> \<Gamma>"
  shows "\<Delta> \<Turnstile> \<tau>"
using prems wf_env_def[of \<Gamma> \<Delta>] by auto

lemma ty_wf_env:
  assumes "\<Gamma>;\<Delta> \<turnstile> e : \<tau>"
  shows "\<Gamma>;\<Delta> \<Turnstile> OK"
using prems by (induct, auto)

lemma ty_wf_env_ty:
  assumes "\<Gamma>;\<Delta> \<turnstile> e : \<tau>"
  shows "\<Delta> \<Turnstile> TY"
proof -
  have "\<Gamma>;\<Delta> \<Turnstile> OK" using ty_wf_env prems by auto
  thus ?thesis using wf_env_wf_env_ty by auto
qed

lemma ok_no_free_db:
  assumes "\<Delta> \<Turnstile> \<tau>"
  shows "free_db_ty \<tau> = []"
using prems proof induct
  case (wf_forall L \<Delta> \<tau>ins)
    hence "finite L" and "finite (dom \<Delta>)" using wf_env_ty_def dom_finite by auto
    then obtain z where Hz: "\<not> z free in type \<tau>ins \<and> \<not> \<Delta> defines z \<and> z \<notin> L"
      using fresh_exists[of "set (free_ty \<tau>ins) \<union> dom \<Delta> \<union> L"] by auto
    hence "set (free_db_ty (freshen_ty z \<tau>ins)) = {}" using wf_forall by auto
    hence "set (free_db_ty \<tau>ins) - {0} = {}" by (auto simp add: vsubst_ty_free_db_ty)
    hence "set (free_db_ty \<tau>ins) \<subseteq> {0}" by auto
    hence "set (shift_db_list (free_db_ty \<tau>ins)) = {}" by (auto simp add: shift_many_sanity)
    hence "shift_db_list (free_db_ty \<tau>ins) = []" by auto
    thus ?case by auto
qed auto

lemma parametrize_ty:
  assumes "\<Delta>,(x:\<star>) \<Turnstile> \<tau>"
  shows "\<exists>\<tau>'. \<not> x free in type \<tau>' \<and> \<tau> = vsubst_ty (TyFreeVar x) i \<tau>'"
proof -
  have "set (free_db_ty \<tau>) = {}" using prems ok_no_free_db by auto
  hence "i \<notin> set (free_db_ty \<tau>)" by auto
  thus ?thesis using parametrize_ty_aux by auto
qed

lemma trivial_case:
  assumes "\<tau>1 \<noteq> \<tau>3"
  and     "\<tau>2 \<noteq> \<tau>3"
  shows "\<not> (if n = i then \<tau>1 else \<tau>2) = \<tau>3"
using prems by (cases "n=i", auto)

lemma subst_var_type_ok:
  assumes "\<Delta>,(x:\<star>) \<Turnstile> vsubst_ty (TyFreeVar x) i \<tau>"
  and "\<not> x free in type \<tau>"
  and "\<Delta> \<Turnstile> \<tau>'"
  and "\<not> \<Delta> defines x"
  shows "\<Delta> \<Turnstile> vsubst_ty \<tau>' i \<tau>"
proof -
  { fix \<Delta>1 \<tau>1
    assume H1: "\<tau>1 = vsubst_ty (TyFreeVar x) i \<tau>"
      and H2: "\<Delta>1 = \<Delta>,(x:\<star>)"
    hence "\<Delta>1 \<Turnstile> \<tau>1" and "\<Delta> \<Turnstile> \<tau>'" and "\<not> \<Delta> defines x" and "\<not> x free in type \<tau>" and "\<Delta> \<Turnstile> TY"
      using prems wf_type_wf_env_ty by auto
    from this H1 H2 have ?thesis proof (induct arbitrary: \<Delta> \<tau> i)
      case (wf_arrow _ \<tau>a' \<tau>r' _ _)
      have "\<Delta>,(x:\<star>) \<Turnstile> \<tau>a'" using prems by auto
      then obtain \<tau>a where Ha: "\<tau>a' = vsubst_ty (TyFreeVar x) i \<tau>a" and Hfa: "\<not> x free in type \<tau>a"
        using parametrize_ty[of x \<Delta> \<tau>a' i] by auto
      hence "\<Delta> \<Turnstile> vsubst_ty \<tau>' i \<tau>a" using prems by blast
      moreover
      have "\<Delta>,(x:\<star>) \<Turnstile> \<tau>r'" using prems by auto
      then obtain \<tau>r where Hr: "\<tau>r' = vsubst_ty (TyFreeVar x) i \<tau>r" and Hfr: "\<not> x free in type \<tau>r"
        using parametrize_ty[of x \<Delta> \<tau>r' i] by auto
      hence "\<Delta> \<Turnstile> vsubst_ty \<tau>' i \<tau>r" using prems by blast
      ultimately
      have "\<Delta> \<Turnstile> vsubst_ty \<tau>' i (\<tau>a \<rightarrow> \<tau>r)" by auto
      moreover
      {
        have "vsubst_ty (TyFreeVar x) i \<tau> = vsubst_ty (TyFreeVar x) i (\<tau>a \<rightarrow> \<tau>r)"
          using prems Ha Hr by auto
        moreover have "\<not> x free in type \<tau>" by assumption
        ultimately
        have "\<tau> = \<tau>a \<rightarrow> \<tau>r" using subst_subst_equal_tyty[of x i \<tau> "\<tau>a\<rightarrow>\<tau>r"] Hfr Hfa by auto
      }
      ultimately
      show ?case using wf_arrow by auto
    next
      case (wf_forall L \<Delta>x \<tau>x \<Delta>y \<tau>y i)
      have "vsubst_ty (TyFreeVar x) i \<tau>y = \<forall>.\<tau>x" using wf_forall by auto
      then obtain \<tau>ins where Heq: "\<tau>y = \<forall>.\<tau>ins" 
        by (cases \<tau>y, auto simp add: trivial_case[of "TyFreeVar x" "\<forall>.\<tau>x"])
      let ?L' = "L \<union> dom \<Delta>x \<union> set (free_ty \<tau>x) \<union> {x}"
      have "\<Delta>y \<Turnstile> \<forall>. vsubst_ty \<tau>' (Suc i) \<tau>ins" proof
        show "\<Delta>y \<Turnstile> TY" using prems by auto
      next
        have "finite (dom \<Delta>x)" using prems wf_env_ty_def[of \<Delta>x] dom_finite[of \<Delta>x] by auto
        thus "finite ?L'" by auto
      next
        {
          fix z
          assume "z \<notin> dom \<Delta>y"
            and "z \<notin> set (free_ty (vsubst_ty \<tau>' (Suc i) \<tau>ins))"
            and "z \<notin> ?L'"
          hence "\<not> \<Delta>x defines z \<and> \<not> z free in type \<tau>x \<and> z \<notin> L" by auto
          hence "(\<forall>xa. xa \<Turnstile> \<tau>' \<longrightarrow>
            x \<notin> dom xa \<longrightarrow>
            (\<forall>xb. x \<notin> set (free_ty xb) \<longrightarrow>
                  xa \<Turnstile> TY \<longrightarrow>
                  (\<forall>xc. freshen_ty z \<tau>x = vsubst_ty (TyFreeVar x) xc xb \<longrightarrow>
                        \<Delta>x,(z:\<star>) = xa,(x:\<star>) \<longrightarrow> xa \<Turnstile> vsubst_ty \<tau>' xc xb)))" using prems by auto 
          moreover
          have "\<Delta>y,(z:\<star>) \<Turnstile> TY" using prems wf_env_ty_def by auto
          moreover
          hence "\<Delta>y,(z:\<star>) \<Turnstile> \<tau>'" using ok_weakening_ty prems by auto
          moreover
          have "\<not> \<Delta>y,(z:\<star>) defines x" using prems by auto
          moreover
          {
            have "\<not> x free in type \<tau>ins" using prems by auto
            moreover
            have "x \<noteq> z" using prems by auto
            ultimately
            have "\<not> x free in type (freshen_ty z \<tau>ins)" by (auto simp add: not_free_subst_ty)
          }
          moreover
          {
            have "\<tau>x = vsubst_ty (TyFreeVar x) (Suc i) \<tau>ins" using prems by auto
            moreover
            have "freshen_ty z (vsubst_ty (TyFreeVar x) (Suc i) \<tau>ins) =
              vsubst_ty (TyFreeVar x) (Suc i) (freshen_ty z \<tau>ins)"
              using subst_subst_ty_swap[of 0 "Suc i"] by auto
            ultimately
            have "freshen_ty z \<tau>x = vsubst_ty (TyFreeVar x) (Suc i) (freshen_ty z \<tau>ins)" by auto
          }
          moreover
          have "\<Delta>x,(z:\<star>) = \<Delta>y,(z:\<star>),(x:\<star>)" using prems by auto
          ultimately
          have "\<Delta>y,(z:\<star>) \<Turnstile> vsubst_ty \<tau>' (Suc i) (freshen_ty z \<tau>ins)" by force
          moreover
          have "free_db_ty \<tau>' = []" using ok_no_free_db prems by auto
          hence "\<forall>k. vsubst_ty (TyFreeVar z) k \<tau>' = \<tau>'" using no_free_db_vars_same_subst_ty by auto
          ultimately
          have "\<Delta>y,(z:\<star>) \<Turnstile> freshen_ty z (vsubst_ty \<tau>' (Suc i) \<tau>ins)"
            using subst_subst_ty_swap[of 0 "Suc i" z \<tau>'] by auto
        }
        thus "\<forall>a. a \<notin> dom \<Delta>y \<and> a \<notin> set (free_ty (vsubst_ty \<tau>' (Suc i) \<tau>ins)) \<and> a \<notin> ?L' \<longrightarrow>
        \<Delta>y,(a:\<star>) \<Turnstile> freshen_ty a (vsubst_ty \<tau>' (Suc i) \<tau>ins)" by auto
      qed
      thus "\<Delta>y \<Turnstile> vsubst_ty \<tau>' i \<tau>y" using Heq by auto
    next
      case wf_ref thus ?case by (cases \<tau>, auto)
    next
      case wf_unit thus ?case by (cases \<tau>, auto)
    next
      case wf_base thus ?case by (cases \<tau>, auto)
    next
      case (wf_tvar \<Delta>y y) thus ?case by (cases \<tau>, auto)
    qed
  }
  thus ?thesis by auto
qed

lemma ty_wf_type:
  assumes "\<Gamma>;\<Delta> \<turnstile> e : \<tau>"
  shows "\<Delta> \<Turnstile> \<tau>"
proof -
  have A: "\<Gamma>;\<Delta> \<Turnstile> OK" using ty_wf_env prems by auto
  have B: "\<Delta> \<Turnstile> TY" using ty_wf_env_ty prems by auto
  from prems A B show ?thesis proof induct
    case t_unit thus ?case by auto
  next
    case t_var thus ?case using wf_env_elim_ty by auto
  next
    case (t_abs L \<Delta> \<Gamma> \<tau>1 \<tau>2 e)
    hence "finite L" and "finite (dom \<Gamma>)" using wf_env_fin by auto
    then obtain z where Hz: "\<not> z free in e \<and> \<not> \<Gamma> defines z \<and> z \<notin> L"
      using fresh_exists[of "set (free_tm e) \<union> dom \<Gamma> \<union> L"] by auto
    moreover
    hence "\<not> \<Gamma> defines z" by auto
    hence "\<Gamma>,(z:\<tau>1);\<Delta> \<Turnstile> OK" using t_abs by (auto simp add: wf_env_weakening)
    ultimately
    have "\<Delta> \<Turnstile> \<tau>2" using t_abs by auto
    thus ?case using prems by auto
  next
    case t_app thus ?case by auto
  next
    case (t_poly L \<Delta> \<Gamma> \<tau> v)      (* make these 10 lines better some time *)
    hence "finite \<Gamma>" by (auto simp add: wf_env_def)
    hence "finite (free_ty_env \<Gamma>)" using free_env_finite by auto
    moreover
    from t_poly have "finite (dom \<Delta>)" using wf_env_ty_fin by auto
    ultimately
    have A: "finite (L \<union> free_ty_env \<Gamma> \<union> dom \<Delta>)" using prems by auto
    show ?case proof
      show "finite (set (free_tmty v) \<union> L \<union> free_ty_env \<Gamma>)" using A by auto
    next
      { fix z
        assume "\<not> \<Delta> defines z"
          and "\<not> z free in type \<tau>"
          and "\<not> z free in term v"
          and "z \<notin> L"
          and "\<not> z free in env \<Gamma>"
        hence "\<not> z free in type \<tau> \<and> z \<notin> (L \<union> free_ty_env \<Gamma> \<union> dom \<Delta>)" by blast
        moreover
        have "\<Gamma>;\<Delta>,(z:\<star>) \<Turnstile> OK" using ok_weakening_tyenv prems by auto
        moreover
        have "\<Delta>,(z:\<star>) \<Turnstile> TY" using prems by (auto simp add: wf_env_ty_def)
        ultimately
        have "\<Delta>,(z:\<star>) \<Turnstile> freshen_ty z \<tau>" using prems by auto
      }
      thus "\<forall>a. a \<notin> dom \<Delta> \<and> a \<notin> set (free_ty \<tau>) \<and> a \<notin> set (free_tmty v) \<union> L \<union> free_ty_env \<Gamma> \<longrightarrow>
        \<Delta>,(a:\<star>) \<Turnstile> freshen_ty a \<tau>" by auto
    qed
  next
    case (t_tyapp \<Delta> \<Gamma> \<tau> \<tau>1 e1)
    have "\<Delta> \<Turnstile> \<forall>.\<tau>1" using prems by auto
    thus ?case proof
      fix L
      assume "\<forall>a. a \<notin> dom \<Delta> \<and> a \<notin> set (free_ty \<tau>1) \<and> a \<notin> L \<longrightarrow>
              \<Delta>,(a:\<star>) \<Turnstile> vsubst_ty (TyFreeVar a) 0 \<tau>1"
        and Hl: "finite L"
      have "finite (dom \<Delta>)" using prems wf_env_ty_fin by auto
      then obtain a where "a \<notin> dom \<Delta> \<and> a \<notin> set (free_ty \<tau>1) \<and> a \<notin> L \<and> \<not> a free in type \<tau>"
        using fresh_exists[of "dom \<Delta> \<union> set (free_ty \<tau>1) \<union> L \<union> set (free_ty \<tau>)"] Hl by auto
      hence "\<Delta>,(a:\<star>) \<Turnstile> vsubst_ty (TyFreeVar a) 0 \<tau>1" using prems by auto
      thus "\<Delta> \<Turnstile> vsubst_ty \<tau> 0 \<tau>1" using subst_var_type_ok[of a \<Delta> 0 \<tau>1 \<tau>] prems by auto
    qed
  next
    case t_loc thus ?case using wf_env_elim_ty by auto
  qed auto
qed

lemma ty_no_free_db:
  assumes "\<Gamma>;\<Delta> \<turnstile> e : \<tau>"
  shows "free_db_tm e = []"
using prems proof induct
  case (t_abs L \<Delta> \<Gamma> \<tau>1 \<tau>2 e_b)
  hence "finite L" and "finite (dom \<Gamma>)" using wf_env_fin by auto
  then obtain z where Hz: "\<not> z free in e_b \<and> \<not> \<Gamma> defines z \<and> z \<notin> L"
    using fresh_exists[of "set (free_tm e_b) \<union> dom \<Gamma> \<union> L"] by auto
  hence "set (free_db_tm (freshen_tm z e_b)) = {}" using t_abs by auto
  hence "set (free_db_tm e_b) - {0} = {}" by (auto simp add: vsubst_tm_free_db_tm)
  hence "set (free_db_tm e_b) \<subseteq> {0}" by auto
  hence "set (shift_db_list (free_db_tm e_b)) = {}" by (auto simp add: shift_many_sanity)
  hence "shift_db_list (free_db_tm e_b) = []" by auto
  thus ?case by auto
next
  case (t_poly L \<Delta> \<Gamma> \<tau> v)   (* make these 10 lines better some time *)
  have "finite L" .
  moreover
  hence "finite \<Gamma>" using wf_env_def t_poly by auto
  moreover
  hence "finite (dom \<Delta>)" using wf_env_ty_fin  t_poly by auto
  ultimately
  obtain z where Hz: "\<not> z free in term v \<and> \<not> z free in type \<tau> \<and> z \<notin> L \<and>
                           \<not> z free in env \<Gamma> \<and> \<not> \<Delta> defines z"
    using t_poly.prems free_env_finite
          fresh_exists[of "set (free_tmty v) \<union> set (free_ty \<tau>) \<union> L \<union> free_ty_env \<Gamma> \<union> dom \<Delta>"]
    by auto
  hence "free_db_tm (freshen_tmty z v) = []" using t_poly.hyps by auto
  thus ?case using vsubst_tmty_free_db by auto
qed auto

lemma ty_no_free_db_ty:
  assumes "\<Gamma>;\<Delta> \<turnstile> e : \<tau>"
  shows "free_db_tmty e = []"
using prems proof induct
  case (t_abs L \<Delta> \<Gamma> \<tau>1 \<tau>2 e_b)
  hence "finite L" and "finite (dom \<Gamma>)" using wf_env_fin by auto
  then obtain z where Hz: "\<not> z free in e_b \<and> \<not> \<Gamma> defines z \<and> z \<notin> L"
    using fresh_exists[of "set (free_tm e_b) \<union> dom \<Gamma> \<union> L"] by auto
  hence "set (free_db_tmty (freshen_tm z e_b)) = {}" using t_abs by auto
  hence "set (free_db_tmty e_b) = {}" by (auto simp add: vsubst_tmty_free_db_tm)
  hence "free_db_tmty e_b = []" by auto
  moreover
  have "free_db_ty \<tau>1 = []" using ok_no_free_db prems by auto
  ultimately
  show ?case by auto
next
  case (t_poly L \<Delta> \<Gamma> \<tau> v)      (* make these 10 lines better some time *)
  have "finite L" .
  moreover
  hence "finite \<Gamma>" using wf_env_def t_poly by auto
  moreover
  hence "finite (dom \<Delta>)" using wf_env_ty_fin t_poly by auto
  ultimately
  obtain z where Hz: "\<not> z free in term v \<and> \<not> z free in type \<tau> \<and> z \<notin> L \<and>
                      \<not> z free in env \<Gamma> \<and> \<not> \<Delta> defines z"
    using t_poly.prems free_env_finite
          fresh_exists[of "set (free_tmty v) \<union> set (free_ty \<tau>) \<union> L \<union> free_ty_env \<Gamma> \<union> dom \<Delta>"]
    by auto
  hence "set (free_db_tmty (freshen_tmty z v)) = {}" using t_poly.hyps by auto
  hence "set (free_db_tmty v) - {0} = {}" by (auto simp add: vsubst_tmty_free_db_tmty)
  hence "set (free_db_tmty v) \<subseteq> {0}" by auto
  hence "set (shift_db_list (free_db_tmty v)) = {}" by (auto simp add: shift_many_sanity)
  hence "shift_db_list (free_db_tmty v) = []" by auto
  thus ?case by auto
next
  case t_tyapp thus ?case using ok_no_free_db by auto
qed auto

lemma ok_same_subst:
  assumes "\<Delta> \<Turnstile> \<tau>"
  shows "vsubst_ty (TyFreeVar a) k \<tau> = \<tau>"
proof -
  have "free_db_ty \<tau> = []" using prems by (auto simp add: ok_no_free_db)
  thus ?thesis by (rule no_free_db_vars_same_subst_ty)
qed

lemma ty_same_subst:
  assumes "\<Gamma>;\<Delta> \<turnstile> e : \<tau>"
  shows "vsubst_tm (TmFreeVar x) k e = e"
proof -
  have "free_db_tm e = []" using prems by (auto simp add: ty_no_free_db)
  thus ?thesis by (rule no_free_db_vars_same_subst)
qed

lemma ty_same_subst_ty:
  assumes "\<Gamma>;\<Delta> \<turnstile> e : \<tau>"
  shows "vsubst_tmty (TyFreeVar a) k e = e"
proof -
  have "free_db_tmty e = []" using prems by (auto simp add: ty_no_free_db_ty)
  thus ?thesis by (rule no_free_db_vars_same_subst_tmty)
qed


lemma ok_weakening_ty_union:
  assumes "\<Delta> \<Turnstile> \<tau>"
  and "(\<Delta>\<union>\<Delta>') \<Turnstile> TY"
  shows "(\<Delta>\<union>\<Delta>') \<Turnstile> \<tau>"
proof -
  have "finite (\<Delta>\<union>\<Delta>')" using prems by (auto simp add: wf_env_ty_def)
  hence "finite \<Delta>'" by auto
  thus ?thesis using prems proof (induct \<Delta>')
    case empty thus ?case by auto
  next
    case (insert p \<Delta>x) thus ?case proof (cases p)
      fix x b
      assume "p = (x, b)"
      thus ?thesis using prems proof (cases b)
        assume "b = Type"
         have "(\<Delta>\<union>\<Delta>x) \<Turnstile> TY" using prems wf_env_strengthening_ty[of x "\<Delta>\<union>\<Delta>x"] by auto
        hence "(\<Delta>\<union>\<Delta>x) \<Turnstile> \<tau>" using prems by auto
        thus ?thesis using ok_weakening_ty prems by auto
      qed
    qed
  qed
qed

lemma ok_strengthening_ty_env:
  assumes "\<Delta>,(a:\<star>) \<Turnstile> \<tau>"
  and Hf: "\<not> a free in type \<tau>"
  shows "\<Delta> \<Turnstile> \<tau>"
proof (cases "(a\<rhd>\<star>) \<in> \<Delta>")
  case False
  { fix \<Delta>'
    assume H: "\<Delta>' = \<Delta>,(a:\<star>)"
    hence H1: "\<Delta>' \<Turnstile> \<tau>" by auto
    hence H2: "\<Delta>' \<Turnstile> TY" by (auto simp add: wf_type_wf_env_ty)
    have "\<Delta>,(a:\<star>) \<Turnstile> TY" using prems by (auto simp add: wf_type_wf_env_ty)
    hence H3: "\<Delta> \<Turnstile> TY" by (rule wf_env_strengthening_ty)
    from H1 H2 H3 False H Hf have ?thesis proof (induct arbitrary: \<Delta>)
      case (wf_arrow \<Delta>x \<tau>1 \<tau>2)
        hence "\<Delta> \<Turnstile> \<tau>1" and "\<Delta> \<Turnstile> \<tau>2" by auto
        thus ?case by auto
    next
      case (wf_forall L \<Delta>x \<tau>ins)
      let ?L = "L \<union> dom \<Delta>x \<union> {a}"
      show ?case proof
        show "\<Delta> \<Turnstile> TY" using wf_forall.prems by auto
      next
        show "finite ?L" using wf_forall.prems 
          by (auto simp add: dom_finite wf_env_ty_def[of \<Delta>])
      next
        { fix z
          assume Hz: "z \<notin> dom \<Delta> \<and> z \<notin> set (free_ty \<tau>ins) \<and> z \<notin> ?L"
          hence "\<Delta>x,(z:\<star>) \<Turnstile> TY" using wf_forall wf_env_ty_def by auto
          hence "\<forall>x. x \<Turnstile> TY \<longrightarrow> (a, Type) \<notin> x \<longrightarrow> \<Delta>x,(z:\<star>) = x,(a:\<star>) \<longrightarrow>
                       a \<notin> set (free_ty (freshen_ty z \<tau>ins)) \<longrightarrow>
                   x \<Turnstile> freshen_ty z \<tau>ins" using wf_forall.hyps Hz by blast
          moreover
          have "a \<notin> set (free_ty (freshen_ty z \<tau>ins))"
            using Hz prems by (auto simp add: not_free_subst_ty)
          moreover
          have "a \<noteq> z" using False prems by auto
          hence "a\<rhd>\<star> \<notin> \<Delta>,(z:\<star>)" using prems by auto
          moreover
          have "\<Delta>x,(z:\<star>) = \<Delta>,(z:\<star>),(a:\<star>)" using prems by auto
          moreover
          hence "\<Delta>,(z:\<star>) \<Turnstile> TY" using wf_forall wf_env_ty_def by auto
          ultimately
          have "\<Delta>,(z:\<star>) \<Turnstile> freshen_ty z \<tau>ins" by auto
        }
        thus "\<forall>z. z \<notin> dom \<Delta> \<and> z \<notin> set (free_ty \<tau>ins) \<and> z \<notin> L \<union> dom \<Delta>x \<union> {a} \<longrightarrow>
                 \<Delta>,(z:\<star>) \<Turnstile> freshen_ty z \<tau>ins" by auto
      qed
    qed auto
  }
  thus ?thesis by auto
next
  case True
  hence "\<Delta> = \<Delta>,(a:\<star>)" by auto
  thus ?thesis using prems by auto
qed

lemma in_env_exists_sanity: (* used only for subst_subst_equal_env proof *)
  fixes \<Gamma> :: tm_env
  assumes "y\<rhd>\<tau> \<in> vsubst_ty_env (TyFreeVar x) i \<Gamma>"
  shows "\<exists>\<tau>'. \<tau> = vsubst_ty (TyFreeVar x) i \<tau>' \<and> y\<rhd>\<tau>' \<in> \<Gamma>"
proof -
  have "(vsubst_ty_env (TyFreeVar x) i \<Gamma>) defines y" using prems by (auto simp add: dom_def)
  hence "\<Gamma> defines y" using vsubst_ty_env_dom_sanity by auto
  then obtain b where "(y, b) \<in> \<Gamma>" using defines_sanity[of y \<Gamma>] by auto
  then obtain \<tau>' where "b = Term \<tau>'" by (cases b)
  thus ?thesis using exI[of _ \<tau>'] using vsubst_ty_env_def prems by auto
qed

lemma subst_subst_equal_env:
  fixes \<Gamma>1 :: tm_env
  fixes \<Gamma>2 :: tm_env
  assumes H: "vsubst_ty_env (TyFreeVar x) i \<Gamma>1 = vsubst_ty_env (TyFreeVar x) i \<Gamma>2"
  and Hx1: "\<not> x free in env \<Gamma>1"
  and Hx2: "\<not> x free in env \<Gamma>2"
  shows "vsubst_ty_env \<tau> i \<Gamma>1 = vsubst_ty_env \<tau> i \<Gamma>2"
proof -
  {
    fix \<Gamma>1 \<Gamma>2
    assume H: "vsubst_ty_env (TyFreeVar x) i \<Gamma>1 = vsubst_ty_env (TyFreeVar x) i \<Gamma>2"
      and Hx1: "\<not> x free in env \<Gamma>1"
      and Hx2: "\<not> x free in env \<Gamma>2"
    { fix y \<tau>'
      assume H': "y\<rhd>\<tau>' \<in> \<Gamma>1"
      hence "(y\<rhd>vsubst_ty (TyFreeVar x) i \<tau>') \<in> vsubst_ty_env (TyFreeVar x) i \<Gamma>1"
        using vsubst_ty_env_def by auto
      hence "(y\<rhd>vsubst_ty (TyFreeVar x) i \<tau>') \<in> vsubst_ty_env (TyFreeVar x) i \<Gamma>2"
        using H by auto
      then obtain \<tau>2 where Heq: "vsubst_ty (TyFreeVar x) i \<tau>' = vsubst_ty (TyFreeVar x) i \<tau>2" 
        and "y\<rhd>\<tau>2 \<in> \<Gamma>2"
        using in_env_exists_sanity[of y "vsubst_ty (TyFreeVar x) i \<tau>'" x i \<Gamma>2] by auto
      moreover
      hence "\<not> x free in type \<tau>2"
        and  "\<not> x free in type \<tau>'"  using Hx2 Hx1 H' free_ty_env_def by auto
      hence "\<tau>' = \<tau>2" using subst_subst_equal_tyty[of x i \<tau>' \<tau>2] Heq by auto
      ultimately
      have "\<exists>\<tau>2. vsubst_ty \<tau> i \<tau>' = vsubst_ty \<tau> i \<tau>2 \<and> y\<rhd>\<tau>2 \<in> \<Gamma>2" using exI[of _ "\<tau>2"] by auto
    }
  }
  hence H1: "\<And>y \<tau>'. y\<rhd>\<tau>' \<in> \<Gamma>1 \<Longrightarrow> \<exists>\<tau>2. vsubst_ty \<tau> i \<tau>' = vsubst_ty \<tau> i \<tau>2 \<and> y\<rhd>\<tau>2 \<in> \<Gamma>2" 
    and H2: "\<And>y \<tau>'. y\<rhd>\<tau>' \<in> \<Gamma>2 \<Longrightarrow> \<exists>\<tau>2. vsubst_ty \<tau> i \<tau>' = vsubst_ty \<tau> i \<tau>2 \<and> y\<rhd>\<tau>2 \<in> \<Gamma>1"
    using prems by auto
  show ?thesis using vsubst_ty_env_def proof auto
    show "\<And>y \<tau>'. y\<rhd>\<tau>' \<in> \<Gamma>1 \<Longrightarrow> \<exists>\<tau>2. vsubst_ty \<tau> i \<tau>' = vsubst_ty \<tau> i \<tau>2 \<and> y\<rhd>\<tau>2 \<in> \<Gamma>2" 
      using H1 by auto
  next
    show "\<And>y \<tau>'. y\<rhd>\<tau>' \<in> \<Gamma>2 \<Longrightarrow> \<exists>\<tau>2. vsubst_ty \<tau> i \<tau>' = vsubst_ty \<tau> i \<tau>2 \<and> y\<rhd>\<tau>2 \<in> \<Gamma>1" 
      using H2 by auto
  qed
qed

lemma ok_subst_equal:
  assumes "\<Delta> \<Turnstile> \<tau>"
  shows "\<tau> = vsubst_ty \<tau>' i \<tau>"
proof -
  have "free_db_ty \<tau> = []" using ok_no_free_db prems by auto
  thus ?thesis using  no_free_db_vars_same_subst_ty by auto
qed

lemma ok_free_env_free_type:
  assumes "\<Delta> \<Turnstile> \<tau>"
  and "\<not> \<Delta> defines a"
  shows "\<not> a free in type \<tau>"
using prems proof induct
  case (wf_tvar \<Delta> b) thus ?case by (cases "a=b", auto simp add: dom_def)
next
  case (wf_forall L \<Delta> \<tau>ins)
  have "finite (dom \<Delta>)" using wf_env_ty_def dom_finite prems by auto
  then obtain b where "b \<notin> dom \<Delta> \<and> b \<notin> set (free_ty \<tau>ins) \<and> b \<notin> L \<and> b \<noteq> a"
    using fresh_exists[of "dom \<Delta> \<union> set (free_ty \<tau>ins) \<union> L \<union> {a}"] prems by blast
  moreover
  hence "\<not> \<Delta>,(b:\<star>) defines a" using prems by auto
  ultimately
  have "\<not> a free in type (freshen_ty b \<tau>ins)" using wf_forall by blast
  hence "\<not> a free in type \<tau>ins" by (unfold freshen_ty) (rule inv_not_free_subst_ty)
  thus ?case by auto
qed auto

lemma ok_free_env_free_env:
  assumes "\<Gamma>;\<Delta> \<Turnstile> OK"
  and "\<not> \<Delta> defines a"
  shows "\<not> a free in env \<Gamma>"
using prems ok_free_env_free_type[of \<Delta> _ a] wf_env_elim_ty by (auto simp add: free_ty_env_def)

lemma ty_not_free_in_term_ty:
  assumes "\<Gamma>;\<Delta>,(x:\<star>) \<turnstile> e : \<tau>"
  and "\<not> x free in term e"
  and "\<not> x free in env \<Gamma>"
  and "\<not> \<Delta> defines x"
  shows "\<not> x free in type \<tau>"
proof -
  from prems show ?thesis proof induct
    case t_var thus ?case using free_ty_env_def by auto
  next
    case (t_abs L \<Delta> \<Gamma> \<tau>ins \<tau>res eins)
    have Hins: "\<not> x free in type \<tau>ins" using t_abs ok_free_env_free_type by auto
    have "finite (dom \<Gamma>)" using ty_wf_env wf_env_fin prems by auto
    then obtain z where "z \<notin> set (free_tm eins) \<and> z \<notin> L \<and> z \<notin> dom \<Gamma>"
      using fresh_exists[of "set (free_tm eins) \<union> L \<union> dom \<Gamma>"] prems by auto
    moreover
    have "\<not> x free in term (freshen_tm z eins)" using prems by (auto simp add: not_free_subst_tm)
    moreover
    have "free_ty_env (\<Gamma>,(z:\<tau>ins)) = free_ty_env \<Gamma> \<union> set (free_ty \<tau>ins)"
      by (auto simp add: free_ty_env_def)
    hence "\<not> x free in env (\<Gamma>,(z:\<tau>ins))" using Hins prems by auto
    ultimately
    have "\<not> x free in type \<tau>res" using t_abs by auto
    thus ?case using Hins by auto
  next
    case t_loc thus ?case using free_ty_env_def by auto
  next
    case (t_tyapp \<Delta> \<Gamma> \<tau> \<tau>1 e1)
    hence "\<not> x free in type \<tau>1" by auto
    moreover
    have "\<not> x free in type \<tau>" using  ok_free_env_free_type prems by auto
    ultimately
    show ?case using not_free_subst_ty by auto
  next
    case (t_poly L \<Delta> \<Gamma> \<tau> v)
    hence "finite (dom \<Delta>)" using wf_env_ty_fin by auto
    moreover
    have "finite (free_ty_env \<Gamma>)" using wf_env_def free_env_finite[of \<Gamma>] t_poly by auto
    ultimately
    obtain a where "\<not> a free in term v" and "\<not> a free in type \<tau>" and "a \<notin> L"
               and "\<not> a free in env \<Gamma>" and "\<not> \<Delta> defines a" and "a \<noteq> x"
      using fresh_exists[of "set (free_tmty v) \<union> set (free_ty \<tau>) \<union> L \<union> free_ty_env \<Gamma>
                              \<union> dom \<Delta> \<union> {x}"] prems by auto
    moreover
    have "\<not> x free in term freshen_tmty a v" using not_free_subst_tmty prems by auto
    ultimately
    have "\<not> x free in type freshen_ty a \<tau>" using t_poly by auto
    hence "\<not> x free in type \<tau>" by (unfold freshen_ty) (rule inv_not_free_subst_ty)
    thus ?case by auto
  qed auto
qed

lemma ok_weakening_ty_env_gen:
  assumes "\<Gamma>;\<Delta> \<Turnstile> OK"
  and "(\<Delta>\<union>\<Delta>') \<Turnstile> TY"
  shows "\<Gamma>;(\<Delta>\<union>\<Delta>') \<Turnstile> OK"
proof -
  have "finite (\<Delta>\<union>\<Delta>')" using prems by (auto simp add: wf_env_ty_def)
  hence "finite \<Delta>'" by auto
  thus ?thesis using prems proof (induct \<Delta>')
    case empty thus ?case by auto
  next
    case (insert p \<Delta>x) thus ?case proof (cases p)
      fix x b
      assume "p = (x, b)"
      thus ?thesis using prems proof (cases b)
        assume "b = Type"
        have "(\<Delta>\<union>\<Delta>x) \<Turnstile> TY" using prems wf_env_strengthening_ty[of x "\<Delta>\<union>\<Delta>x"] by auto
        hence "\<Gamma>;(\<Delta>\<union>\<Delta>x) \<Turnstile> OK" using prems by auto
        thus ?thesis proof (cases "p \<in> \<Delta>x")
          case True
          hence "\<Delta>x = insert p \<Delta>x" by auto
          thus ?thesis using prems by auto
        next
          case False
          thus ?thesis proof (cases "p \<in> \<Delta>")
            case True
            hence "\<Delta> = insert p \<Delta>" by auto
            hence "(\<Delta> \<union> insert p \<Delta>x) = \<Delta>\<union>\<Delta>x" by auto
            thus ?thesis using prems by auto
          next
            case False
            have Hok: "(\<Delta>\<union>\<Delta>x) \<Turnstile> TY" using prems wf_env_strengthening_ty[of x "\<Delta>\<union>\<Delta>x"] by auto
            have "\<not> (\<Delta>\<union>\<Delta>x) defines x" proof
              assume "x \<in> dom (\<Delta>\<union>\<Delta>x)"
              then obtain b where "(x, b) \<in> (\<Delta>\<union>\<Delta>x)" by (auto simp add: dom_def)
              show False using prems proof (cases b)
                assume "b = Type"
                thus ?thesis using prems by auto
              qed
            qed
            moreover
            hence "\<Gamma>;(\<Delta>\<union>\<Delta>x) \<Turnstile> OK" using prems wf_env_strengthening_ty[of x "\<Delta>\<union>\<Delta>x"] Hok
              by auto
            ultimately
            show ?thesis using ok_weakening_tyenv[of _ "\<Delta>\<union>\<Delta>x" x] prems by auto
          qed
        qed
      qed
    qed
  qed
qed

lemma subst_convert_ty:
  assumes "\<Delta> \<Turnstile> \<tau>''"
  and "\<tau>'' = vsubst_ty (TyFreeVar x) i \<tau>"
  and "\<Delta> \<Turnstile> \<tau>'"
  shows "vsubst_ty \<tau>' i \<tau> = vsubst_ty \<tau>' j (convert_ty j i \<tau>)"
using prems proof (induct arbitrary: \<tau> i j)
  case (wf_unit \<Delta> \<tau> i j) thus ?case by (cases \<tau>, auto)
next
  case (wf_base \<Delta> n \<tau> i j) thus ?case by (cases \<tau>, auto)
next
  case (wf_tvar \<Delta> xa \<tau> i j) thus ?case by (cases \<tau>, auto)
next
  case (wf_arrow \<Delta> \<tau>1 \<tau>2 \<tau> i j) thus ?case by (cases \<tau>, auto)
next
  case (wf_ref \<Delta> \<tau>ins \<tau> i j) thus ?case by (cases \<tau>, auto)
next
  case (wf_forall L \<Delta> \<tau>ins \<tau> i j) thus ?case proof (cases \<tau>, auto)
    fix \<tau>ins2
    assume H: "\<forall>a. a \<notin> dom \<Delta> \<and>
              a \<notin> set (free_ty (vsubst_ty (TyFreeVar x) (Suc i) \<tau>ins2)) \<and>
              a \<notin> L \<longrightarrow>
              (\<forall>xa xb.
                  vsubst_ty (TyFreeVar a) 0
                   (vsubst_ty (TyFreeVar x) (Suc i) \<tau>ins2) =
                  vsubst_ty (TyFreeVar x) xb xa \<longrightarrow>
                  \<Delta>,(a:\<star>) \<Turnstile> \<tau>' \<longrightarrow>
                  (\<forall>x. vsubst_ty \<tau>' xb xa =
                       vsubst_ty \<tau>' x (convert_ty x xb xa)))"
    moreover
    have "finite (dom \<Delta>)" and "finite L" 
      using prems by (auto simp add: wf_env_ty_def dom_finite)
    then obtain a
      where Ha: "a \<notin> dom \<Delta> \<and> \<not> a free in type vsubst_ty (TyFreeVar x) (Suc i) \<tau>ins2 \<and>
                 \<not> a free in type \<tau>' \<and> a \<notin> L"
      using fresh_exists[of "dom \<Delta> \<union> set (free_ty (vsubst_ty (TyFreeVar x) (Suc i) \<tau>ins2)) \<union>
                             set (free_ty \<tau>') \<union> L"]
      by auto
    moreover
    have "vsubst_ty (TyFreeVar a) 0 (vsubst_ty (TyFreeVar x) (Suc i) \<tau>ins2) =
                  vsubst_ty (TyFreeVar x) (Suc i) (vsubst_ty (TyFreeVar a) 0 \<tau>ins2)"
      using subst_subst_ty_swap[of "Suc i" 0] by auto
    moreover
    have "\<Delta>,(a:\<star>) \<Turnstile> \<tau>'" using prems by (auto simp add: ok_weakening_ty)
    ultimately
    have "vsubst_ty \<tau>' (Suc i) (freshen_ty a \<tau>ins2) =
      vsubst_ty \<tau>' (Suc j) (convert_ty (Suc j) (Suc i) (freshen_ty a \<tau>ins2))"
      by auto
    moreover
    have Hk: "\<forall>k. vsubst_ty (TyFreeVar a) k \<tau>' = \<tau>'"
      using prems by (auto simp add: ok_same_subst)
    hence "freshen_ty a (vsubst_ty \<tau>' (Suc i) \<tau>ins2) = vsubst_ty \<tau>' (Suc i) (freshen_ty a \<tau>ins2)"
      using subst_subst_ty_swap[of 0 "Suc i"] by auto
    moreover
    have "freshen_ty a (vsubst_ty \<tau>' (Suc j) (convert_ty (Suc j) (Suc i) \<tau>ins2)) =
           vsubst_ty \<tau>' (Suc j) (convert_ty (Suc j) (Suc i) (freshen_ty a \<tau>ins2))"
      using Hk subst_convert_subst_swap by auto
    ultimately
    have "freshen_ty a (vsubst_ty \<tau>' (Suc i) \<tau>ins2) =
            freshen_ty a (vsubst_ty \<tau>' (Suc j) (convert_ty (Suc j) (Suc i) \<tau>ins2))" by auto
    moreover
    have "\<not> a free in type vsubst_ty (TyFreeVar x) (Suc i) \<tau>ins2" using prems by auto
    hence H\<tau>ins2: "\<not> a free in type \<tau>ins2" by (rule inv_not_free_subst_ty)
    hence "\<not> a free in type (vsubst_ty \<tau>' (Suc i) \<tau>ins2)"
      using prems by (auto simp add: not_free_subst_ty)
    moreover
    from H\<tau>ins2 have "\<not> a free in type convert_ty (Suc j) (Suc i) \<tau>ins2"
      by (rule not_free_convert_ty)
    hence "\<not> a free in type vsubst_ty \<tau>' (Suc j) (convert_ty (Suc j) (Suc i) \<tau>ins2)"
      using prems by (auto simp add: not_free_subst_ty)
    ultimately
    show "vsubst_ty \<tau>' (Suc i) \<tau>ins2 =
            vsubst_ty \<tau>' (Suc j) (convert_ty (Suc j) (Suc i) \<tau>ins2)"
      using subst_subst_equal_tyty[of a 0 "vsubst_ty \<tau>' (Suc i) \<tau>ins2"
        "vsubst_ty \<tau>' (Suc j) (convert_ty (Suc j) (Suc i) \<tau>ins2)"] by auto
  qed
qed

lemma wf_env_no_free_db:
  assumes "\<Gamma>;\<Delta> \<Turnstile> OK"
  shows "free_db_ty_env \<Gamma> = {}"
proof (auto simp add: free_db_ty_env_def)
  fix x \<tau>
  assume "(x\<rhd>\<tau>) \<in> \<Gamma>"
  have "\<Delta> \<Turnstile> \<tau>" using prems by (auto simp add: wf_ty_ok_ty)
  thus "free_db_ty \<tau> = []" by (rule ok_no_free_db)
qed

lemma no_free_db_vars_same_subst_ty_env:
  assumes "free_db_ty_env \<Gamma> = {}"
  shows "vsubst_ty_env \<tau> k \<Gamma> = \<Gamma>"
proof (auto simp add: vsubst_ty_env_def)
  fix x \<tau>'
  assume "(x\<rhd>\<tau>') \<in> \<Gamma>"
  have "free_db_ty \<tau>' = []" using prems by (auto simp add: free_db_ty_env_def)
  hence "vsubst_ty \<tau> k \<tau>' = \<tau>'" by (rule no_free_db_vars_same_subst_ty)
  thus "(x\<rhd>(vsubst_ty \<tau> k \<tau>')) \<in> \<Gamma>" by auto
next
  fix x b
  assume "(x, b) \<in> \<Gamma>"
  then obtain \<tau>' where "b = Term \<tau>'" by (cases b)
  have "free_db_ty \<tau>' = []" using prems by (auto simp add: free_db_ty_env_def)
  hence "vsubst_ty \<tau> k \<tau>' = \<tau>'" by (rule no_free_db_vars_same_subst_ty)
  hence "b = Term (vsubst_ty \<tau> k \<tau>')" and "(x\<rhd>\<tau>') \<in> \<Gamma>" using prems by auto
  thus "\<exists>\<tau>'. b = Term (vsubst_ty \<tau> k \<tau>') \<and> (x\<rhd>\<tau>') \<in> \<Gamma>" by auto
qed

lemma subst_convert_env:
  assumes "\<Gamma>';\<Delta> \<Turnstile> OK"
  and "\<Gamma>' = vsubst_ty_env (TyFreeVar x) i \<Gamma>"
  and "\<Delta> \<Turnstile> \<tau>'"
  shows "vsubst_ty_env \<tau>' i \<Gamma> = vsubst_ty_env \<tau>' j (convert_ty_env j i \<Gamma>)"
proof (auto simp add: vsubst_ty_env_def convert_ty_env_def)
  fix y \<tau>
  assume "(y\<rhd>\<tau>) \<in> \<Gamma>"
  hence "(y\<rhd>(vsubst_ty (TyFreeVar x) i \<tau>)) \<in> \<Gamma>'" using prems by (auto simp add: vsubst_ty_env_def)
  hence "\<Delta> \<Turnstile> vsubst_ty (TyFreeVar x) i \<tau>" using prems by (auto simp add: wf_ty_ok_ty)
  hence "vsubst_ty \<tau>' i \<tau> = vsubst_ty \<tau>' j (convert_ty j i \<tau>)" and
    "(y\<rhd>\<tau>) \<in> \<Gamma>" using prems by (auto simp add: subst_convert_ty)
  thus "\<exists>\<tau>''. vsubst_ty \<tau>' i \<tau> = vsubst_ty \<tau>' j \<tau>'' \<and> (\<exists>\<tau>. \<tau>'' = convert_ty j i \<tau> \<and> (y\<rhd>\<tau>) \<in> \<Gamma>)"
    by auto
next
  fix y \<tau>
  assume "(y\<rhd>\<tau>) \<in> \<Gamma>"
  hence "(y\<rhd>(vsubst_ty (TyFreeVar x) i \<tau>)) \<in> \<Gamma>'" using prems by (auto simp add: vsubst_ty_env_def)
  hence "\<Delta> \<Turnstile> vsubst_ty (TyFreeVar x) i \<tau>" using prems by (auto simp add: wf_ty_ok_ty)
  hence "vsubst_ty \<tau>' i \<tau> = vsubst_ty \<tau>' j (convert_ty j i \<tau>)"
    using prems by (auto simp add: subst_convert_ty)
  hence "vsubst_ty \<tau>' j (convert_ty j i \<tau>) = vsubst_ty \<tau>' i \<tau>" and "(y\<rhd>\<tau>) \<in> \<Gamma>"
    using prems by auto
  thus "\<exists>\<tau>''. vsubst_ty \<tau>' j (convert_ty j i \<tau>) = vsubst_ty \<tau>' i \<tau>'' \<and> (y\<rhd>\<tau>'') \<in> \<Gamma>" by auto
qed

lemma uniq_typing_in_env:
  assumes "\<Gamma>;\<Delta> \<Turnstile> OK"
  and "(x\<rhd>\<tau>) \<in> \<Gamma>"
  and "(x\<rhd>\<tau>') \<in> \<Gamma>"
  shows "\<tau>=\<tau>'"
using prems by (auto simp add: wf_env_def)

lemma not_free_convert_type:
  assumes "\<not> x free in type \<tau>"
  shows "\<not> x free in type (convert_ty i j \<tau>)"
using prems by (induct \<tau> arbitrary: i j, auto)

lemma not_free_convert_env:
  assumes "\<not> x free in env \<Gamma>"
  shows "\<not> x free in env (convert_ty_env i j \<Gamma>)"
proof (auto simp add: free_ty_env_def convert_ty_env_def)
  fix y \<tau>
  assume "x free in type convert_ty i j \<tau>"
    and "(y\<rhd>\<tau>) \<in> \<Gamma>"
  hence "\<not> x free in type \<tau>" using prems by (auto simp add: free_ty_env_def)
  hence "\<not> x free in type convert_ty i j \<tau>" by (rule not_free_convert_type)
  thus False by contradiction
qed

lemma not_free_subst_env:
  assumes "\<not> a free in env \<Gamma>"
  and "\<not> a free in type \<tau>"
  shows "\<not> a free in env vsubst_ty_env \<tau> i \<Gamma>"
proof (auto simp add: free_ty_env_def)
  fix y \<tau>'
  assume "a free in type \<tau>'"
    and "(y\<rhd>\<tau>') \<in> vsubst_ty_env \<tau> i \<Gamma>"
  then obtain \<tau>'' where "(y\<rhd>\<tau>'') \<in> \<Gamma>" and "\<tau>' = vsubst_ty \<tau> i \<tau>''"
    by (auto simp add: vsubst_ty_env_def)
  hence "\<not> a free in type \<tau>''" using prems by (auto simp add: free_ty_env_def)
  hence "\<not> a free in type \<tau>'" using prems by (auto simp add: not_free_subst_ty)
  thus False by contradiction
qed

lemma free_subst_env:
  assumes "a free in env \<Gamma>"
  shows "a free in env vsubst_ty_env \<tau> i \<Gamma>"
using prems proof (auto simp add: free_ty_env_def vsubst_ty_env_def)
  fix x \<tau>'
  assume "a free in type \<tau>'"
    and "(x\<rhd>\<tau>') \<in> \<Gamma>"
  hence "a free in type vsubst_ty \<tau> i \<tau>'" by (auto simp add: free_subst_ty)
  thus "\<exists>A. (\<exists>y \<tau>y. A = set (free_ty \<tau>y) \<and> (\<exists>\<tau>''. \<tau>y = vsubst_ty \<tau> i \<tau>'' \<and> (y\<rhd>\<tau>'') \<in> \<Gamma>)) \<and> a \<in> A"
    using exI[of _ "set (free_ty (vsubst_ty \<tau> i \<tau>'))"] exI[of _ x] exI[of _ "vsubst_ty \<tau> i \<tau>'"]
      prems by auto
qed

lemma inv_not_free_subst_env:
  assumes "\<not> a free in env vsubst_ty_env \<tau> i \<Gamma>"
  shows "\<not> a free in env \<Gamma>"
using prems free_subst_env by auto

lemma wf_env_subst_eq:
  assumes "\<Gamma>;\<Delta> \<Turnstile> OK"
  shows "\<Gamma> = vsubst_ty_env \<tau>' i \<Gamma>"
proof (auto simp add: vsubst_ty_env_def)
  fix x b
  assume "(x, b) \<in> \<Gamma>"
  then obtain \<tau> where "b = Term \<tau>" by (cases b)
  hence "\<Delta> \<Turnstile> \<tau>" using prems by (auto simp add: wf_ty_ok_ty)
  hence "\<tau> = vsubst_ty \<tau>' i \<tau>" by (rule ok_subst_equal)
  thus "\<exists>\<tau>. b = Term (vsubst_ty \<tau>' i \<tau>) \<and> (x\<rhd>\<tau>) \<in> \<Gamma>" using prems by auto
next
  fix x \<tau>
  assume "(x\<rhd>\<tau>) \<in> \<Gamma>"
  hence "\<Delta> \<Turnstile> \<tau>" using prems by (auto simp add: wf_ty_ok_ty)
  hence "\<tau> = vsubst_ty \<tau>' i \<tau>" by (rule ok_subst_equal)
  thus "(x\<rhd>(vsubst_ty \<tau>' i \<tau>)) \<in> \<Gamma>" by auto
qed

end
